/***************************************************************************
                          cdemu.c  -  description
                             -------------------
    begin                : 2002-12-29
    copyright            : 2002-2004 by Robert Penz <robert.penz@outertech.com>
                           2003-2004 by Mike Frysinger <vapier@netzero.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
 *                                                                         *
 ***************************************************************************/

/*
 * Version 0.6 - 2003-10-13
 * add linux-2.6.x support and break up code to try to
 * blur the differences between 2.4.x and 2.6.x
 *
 * Version 0.5 - 2003-07-02
 * rename to CDemu
 * rewrite of major parts to be 100% compatible to a cdrom
 * user space interface changed for new userspace program

 * Version 0.4 - 2oo3-04-08
 * removed sys_call_table use
 * fixed bug with devfsd
 * some extra error handling for the userspace program

 * Version 0.3 - 2oo3-01-11
 * added check if a "cd is loaded"
 * removed all warnings
 * code cleanup

 * Version 0.2 - 2oo3-01-12
 * support for proc fs
 * changed the vclib.c so it can be used by gui programs

 * Version 0.1 - 2oo3-01-10
 * first test release (data and raw mode work - but no checks, if you do something
 * wrong you'll kill the kernel ;-)
 *
 * gcc cdemu.c -o cdemu.o -c -I/usr/src/linux/include -O2 -D__KERNEL__ -DMODULE -Wall
 */

#include "cdemu_kernel.h"
#include "cdemu.h"

/* *********************************************************************** */
/*                        helper functions (generic kernel)                */
/* *********************************************************************** */

static inline int cdemu_msf_2_sector(int minute, int second, int frame)
{
	return frame + (second + minute * 60 ) * 75;
}

static inline int cdemu_mode_2_sector_size(int mode)
{
	switch (mode) {
		case AUDIO:			return CD_FRAMESIZE_RAW;
		case MODE1_2352:	return CD_FRAMESIZE_RAW;
		case MODE1_2048:	return CD_FRAMESIZE;
		case MODE2_2352:	return CD_FRAMESIZE_RAW;
		case MODE2_2336:	return CD_FRAMESIZE_RAW0;
		default:			return 0;
	}
}

static void convert_track_mode(struct cdrom_tocentry *ep, int mode)
{
	ep->cdte_datamode = 0;
	switch (mode) {
		case AUDIO:			ep->cdte_ctrl = 0;
							return;
		case MODE1_2352:	ep->cdte_ctrl = ep->cdte_ctrl | CDROM_DATA_TRACK;
							return;
		case MODE1_2048:	ep->cdte_ctrl = ep->cdte_ctrl | CDROM_DATA_TRACK;
							return;
		case MODE2_2352:	ep->cdte_ctrl = ep->cdte_ctrl | CDROM_DATA_TRACK;
							return;
		case MODE2_2336:	ep->cdte_ctrl = ep->cdte_ctrl | CDROM_DATA_TRACK;
							return;
		default:			return;
	}
}

static int cdemu_findtrack(struct cdemu_device *vc, int sector)
{
	int i;
	for(i = vc->num_tracks - 1; i >= 0; i--)
		if (vc->tracks[i].start_sector <= sector)
			return i;
	return 0;
}

/* return 0 upon success */
int read_toc_header(struct cdrom_tochdr *hp, struct cdemu_device *vc)
{
	if (vc->loaded == 0)
		printk_debug("read_toc_header: here is no cd loaded.\n");
	if (hp) {
		hp->cdth_trk0 = 1;
		hp->cdth_trk1 = vc->num_tracks ;
		return 0;
	}
	return -1;
}

/* cdrom.c guarantees that cdte_format == CDROM_MSF */
void get_toc_entry(struct cdrom_tocentry *ep, struct cdemu_device *vc)
{
	tTrack *aTrack;

	if (vc->loaded == 0)
		printk_debug("get_toc_entry: here is no cd loaded\n");

	/* lead out or in our boundies, everthing els is an erro*/
	if (ep->cdte_track == CDROM_LEADOUT)
		aTrack = &(vc->tracks[vc->num_tracks]);
	else if (ep->cdte_track > vc->num_tracks  || ep->cdte_track < 1) {
		printk_debug("get_toc_entry: there is no track %i\n", ep->cdte_track);
		return;
	} else /* normal case */
		aTrack = &(vc->tracks[ep->cdte_track]) - 1;

	ep->cdte_addr.msf.frame = aTrack->frame;
	ep->cdte_addr.msf.second = aTrack->second;
	ep->cdte_addr.msf.minute = aTrack->minute;

	convert_track_mode(ep, aTrack->mode);

	printk_debug("get_toc_entry: track %i, msf %i:%i%i, mode %i\n",
			ep->cdte_track,
			aTrack->minute, aTrack->second, aTrack->frame,
			aTrack->mode);
}

static int cdemu_read_actor(read_descriptor_t *desc, struct page *page, unsigned long offset, unsigned long size)
{
	char *kaddr;
	unsigned long count = desc->count;

	if (size > count)
		size = count;

	kaddr = kmap(page);
	memcpy(desc->buf, kaddr + offset, size);
	kunmap(page);

	desc->count = count - size;
	desc->written += size;
	desc->buf += size;
	return size;
}

static int cdemu_transfer(struct cdemu_device *vc, const struct request *req)
{
	ssize_t retval;
	int startposition, track;
	int offset;
	read_descriptor_t desc;
	loff_t position;
	long sector = req->sector / 4;
	/* is only != 0 if the sector is not inline with our sectors
	vmware does request such sectors ... mount does not */
	int sec_offset = req->sector % 4;

	/* thats just testing code as I don't know really how it works */

	track = cdemu_findtrack(vc, sector);

	/* we know the track get the mode of it */
	switch (vc->tracks[track].mode) {
		case MODE1_2352:
			offset = CD_HEAD_SIZE + CD_SYNC_SIZE;
			break;
		case MODE1_2048:
			offset = 0;
			break;
		default:
		{
			static int count = 0;
			if (count++ < 5)			/* limit output */
				printk_cdemu(KERN_WARNING, "Normal read not possible for track mode '%i'\n", vc->tracks[track].mode);
			return 0;
		}
	}

	/* get the mode of the bin file part and calc the positon */
	startposition = vc->tracks[track].start_offset +
	                (sector - vc->tracks[track].start_sector)
	                * cdemu_mode_2_sector_size(vc->tracks[track].mode);

	/* we only support read, and we told the kernel that, so we won't check it here */

	position = startposition + offset + sec_offset * 512;
	desc.written = 0;
	desc.count = 512 * req->current_nr_sectors;
	desc.buf = req->buffer;
	desc.error = 0;
	printk_debug("cdemu_transfer: do_generic_file_read(position=%08x, count=%08x)\n", (unsigned int) position, desc.count);
	do_generic_file_read(vc->backing_file, &position, &desc, cdemu_read_actor);

	retval = desc.written;

	if (!retval) {
		retval = desc.error;
		printk_debug("cdemu_transfer: retval error %i\n", retval);
		return retval;
	}

	return 1;
}

int unload_cd(struct cdemu_device *vc)
{
	printk_debug("unload_cd: unlocking start\n");

	/* check if the cdrom is locked */
	if (vc->locked)
		return -EBUSY;

	/* make sure uid is ok */
	if (current->uid != 0 || vc->uid != current->uid)
		return -EACCES;

	/* just close the bin file */
	vc->loaded = 0;
	fput(vc->backing_file);
	DOWN_MODULE_USE();

	printk_debug("unload_cd: unlocking end\n");
	return 0;
}

/* *********************************************************************** */
/*                        helper functions (specific kernel)               */
/* *********************************************************************** */

static struct cdemu_device *cdemu_locate_device(const struct request *req)
{
	struct cdemu_device *vc;
#ifdef IS_LINUX_2_4
	int dev;
#endif

	printk_debug("cdemu_locate_device: start\n");

	/* try and locate the right cdemu */
#ifdef IS_LINUX_2_6
	if (!req || !req->rq_disk || !req->rq_disk->private_data)
#else
	dev = DEVICE_NR(req->rq_dev);
	if (dev >= max_cdemus)
#endif
		return NULL;

#ifdef IS_LINUX_2_6
	vc = (struct cdemu_device*)req->rq_disk->private_data;
#else
	vc = &cdemu_dev[dev];
#endif
	printk_debug("cdemu_locate_device: found %s\n", vc->info.name);
	return vc;
}

/* *********************************************************************** */
/*                        device stuff                                     */
/* *********************************************************************** */

static void cdemu_request(request_queue_t *q)
{
	struct cdemu_device *vc;
	struct request *req;
	int status;
//	static int count = 0;

	while (1) {
#ifdef IS_LINUX_2_6
		req = elv_next_request(q);			/* returns when queue is empty */
#else
		INIT_REQUEST;
		req = CURRENT;
#endif

		if (req == NULL)
			return;

#ifdef IS_LINUX_2_6
		spin_unlock_irq(q->queue_lock);
#endif

		/* Which "device" are we using? */
		vc = cdemu_locate_device(req);
		if (vc == NULL) {
//			if (count++ < 5)			/* limit output */
			printk_cdemu(KERN_WARNING, "Request for non-existant device\n");
			status = 0;
			goto end_this_request;
		}
		/* check if we have a "cd loaded" */
		if (!vc->loaded) {
			printk_cdemu(KERN_WARNING, "Request when drive #%i isn't loaded\n", vc->number);
			status = -ENOMEDIUM;
			goto end_this_request;
		}

		/* Perform the transfer, device seems to be good */
		status = cdemu_transfer(vc, req);

		/* now lets end the request */
end_this_request:
#ifdef IS_LINUX_2_6
		spin_lock_irq(q->queue_lock);
		end_request(req, status);
#else
		end_request(status);
#endif
	}
}

/*                        interface to the user space */
static int cdemu_open(struct cdrom_device_info *cdi, int purpose)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_open: start %s\n", cdi->name);

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		printk_debug("cdemu_open: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	/* one more and lock it */
	vc->refcnt++;

	/* one use more and lock */
	UP_MODULE_USE();

	printk_debug("cdemu_open: end, %i still in use\n", vc->refcnt);
	return 0;
}

/*                        interface to the user space */
static void cdemu_release(struct cdrom_device_info *cdi)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_release: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return;
	vc = cdi->handle;

	/* one less and unlock it */
	vc->refcnt--;

	/* one use less and unlock */
	DOWN_MODULE_USE();

	printk_debug("cdemu_release: end, %i still in use\n", vc->refcnt);
	return;
}

static int cdemu_drive_status(struct cdrom_device_info *cdi, int slot_nr)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_drive_status: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		printk_debug("cdemu_drive_status: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	printk_debug("cdemu_drive_status: loaded: %i\n", vc->loaded);
	return (vc->loaded) ? CDS_DISC_OK : CDS_NO_DISC;
}

static int cdemu_media_changed(struct cdrom_device_info *cdi, int disc_nr)
{
	struct cdemu_device *vc;
	int ret;

	printk_debug("cdemu_media_changed: start %s\n", cdi->name);

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		printk_debug("cdemu_media_changed: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	ret = vc->media_changed;
	vc->media_changed = 0;  /* clear bit */

	printk_debug("cdemu_media_changed: end\n");
	return ret;
}

static int cdemu_tray_move(struct cdrom_device_info *cdi, int position)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_tray_move: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	/* 1: eject */
	if (position)
		if (vc->loaded)
			unload_cd(vc);

	printk_debug("cdemu_tray_move: end\n");
	return 0;
}

/* locks or unlocks door lock==1: lock; return 0 upon success */
static int cdemu_lock_door(struct cdrom_device_info *cdi, int lock)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_lock_door: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	vc->locked = lock;

	return 0;
}

/* no idea how I should implement that .. so just say we set it */
static int cdemu_select_speed(struct cdrom_device_info *cdi, int speed)
{
	printk_debug("cdemu_select_speed: start\n");
	return 1;
}

/* bin/cue files can not be in mutlisession*/
static int cdemu_get_last_session(struct cdrom_device_info *cdi, struct cdrom_multisession *mssp)
{
	printk_debug("cdemu_get_last_session: start\n");

	/* the cd is not multi session */
	mssp->xa_flag = 0;

	return 0;
}

static int cdemu_reset(struct cdrom_device_info *cdi)
{
	int i;

	printk_debug("cdemu_reset: start\n");

	for (i = 0; i < max_cdemus; ++i)
		unload_cd(&cdemu_dev[i]);

	return 0;
}

static int cdemu_audio_ioctl(struct cdrom_device_info *cdi, unsigned int cmd, void *arg)
{
	struct cdemu_device *vc;

	printk_debug("cdemu_audio_ioctl: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle) {
		printk_debug("cdemu_audio_ioctl: -ENODEV\n");
		return -ENODEV;
	}
	vc = cdi->handle;

	switch (cmd) {
		case CDROMREADTOCHDR:
			return read_toc_header((struct cdrom_tochdr *) arg, vc);
		case CDROMREADTOCENTRY:
			get_toc_entry((struct cdrom_tocentry *) arg, vc);
			return 0;
		case CDROMPLAYMSF:
		case CDROMPLAYTRKIND:
		case CDROMSTOP:
		case CDROMPAUSE:
		case CDROMRESUME:
		case CDROMSTART:
		case CDROMVOLCTRL:
			return 0;
		case CDROMSUBCHNL:
			return -EINVAL; 
		default:
			return -EINVAL;
	}
}

static int cdemu_ioctl(struct cdrom_device_info *cdi, unsigned int cmd, unsigned long arg)
{
	struct cdemu_device *vc;
	struct cdrom_msf msf;
	struct cdemu_info cdemu_info;

	loff_t position;
	int sector;
	int track;
	int tmp;

	printk_debug("cdemu_ioctl: start\n");

	/* lets find the device and call it vc for easier use */
	if (!cdi || !cdi->handle)
		return -ENODEV;
	vc = cdi->handle;

	if (cmd != CDROMREADRAW)
		printk_debug("cdemu_ioctl: start - command:%x at cdrom %i\n", cmd, vc->number);

	/* check if we've a "cd loaded" and we're not trying to load one */
	if (vc->loaded == 0 && cmd != CDEMU_LOADCD) {
		if (cmd != CDROMREADRAW)
			printk_debug("cdemu_ioctl: here is no cd loaded.\n");
		return -ENOMEDIUM;
	}

	switch (cmd) {
		case CDROMREADCOOKED:	/*read data in mode 1 (2048 Bytes) */
		case CDROMREADRAW:		/*read data in mode 2 (2336 Bytes) */
			if (copy_from_user(&msf, (void *) arg, sizeof(msf)))
				return -EFAULT;

			sector = cdemu_msf_2_sector(msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0);
			track = cdemu_findtrack(vc, sector);

			/* get the mode of the bin file part and calc the positon */
			position = vc->tracks[track].start_offset + 
					(sector - vc->tracks[track].start_sector)
					* cdemu_mode_2_sector_size(vc->tracks[track].mode);

			printk_debug("sector:%i track:%i position %i (%i:%i:%i)\n",
					sector, track, (int)position,
					msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0);

			/* check if the track is at its end*/
			if (position >= vc->tracks[track+1].start_offset) {
				printk_debug("cdemu_ioctl: position out of track - sector:%i track:%i position %i\n",
						sector, track, (int)position);
				return -EFAULT;
			}

			if (cmd == CDROMREADRAW) {
				tmp = generic_file_read(vc->backing_file, (void*)arg, CD_FRAMESIZE_RAW, &position);
				if (tmp != CD_FRAMESIZE_RAW)
					printk_debug("only read %i bytes\n",tmp);
			} else {	/*CDROMREADCOOKED*/
				tmp = generic_file_read(vc->backing_file, (void*)arg, CD_FRAMESIZE, &position);
				if (tmp != CD_FRAMESIZE)
					printk_debug("only read %i bytes\n",tmp);
			}

			return 0;

		case CDROMSEEK:			/* seek msf address - not supported */
			return 0;

		case CDROMREADMODE1:	/* set read data in mode 1 */
		case CDROMREADMODE2:	/* set read data in mode 2 */
			printk_debug("cdemu_ioctl: not supported that CDROMREAD: %x\n", cmd);
			return 0;

		case CDEMU_LOADCD:
			printk_debug("cdemu_ioctl: loading start\n");
			/* check if the cdrom is locked */
			if (vc->locked)
				return -EBUSY;

			/* check if the cdrom is loaded already */
			if (vc->loaded)
				return -EBUSY;

			/* just update our virtual cd data */
			if (copy_from_user(&cdemu_info, (void *)arg, sizeof(cdemu_info)))
				return -EFAULT;

			vc->backing_file = fget(cdemu_info.ffd);
			if (!vc->backing_file)
				return -EBADF;

			memcpy(vc->tracks, cdemu_info.tracks, sizeof(cdemu_info.tracks));
			vc->num_tracks = cdemu_info.num_tracks;
			strncpy(vc->comment, cdemu_info.comment, CDEMU_COMMENT_SIZE);
			vc->comment[CDEMU_COMMENT_SIZE-1] = '\0';
			vc->uid = current->uid;
			printk_cdemu(KERN_INFO, "loaded %s cd (%i track[s]) on drive %s by uid %i\n",
			             vc->comment, vc->num_tracks, cdi->name, vc->uid);

			/* hard coded to get the size of the first track */
			vc->size = vc->tracks[1].start_sector * CD_FRAMESIZE;

			/* only inc use count if there was no cd in the drive */
			if (!vc->loaded)
				UP_MODULE_USE();

			vc->loaded = 1;

			printk_debug("cdemu_ioctl: loading end\n");

			return 0;

		case CDEMU_UNLOADCD:
			printk_cdemu(KERN_INFO, "unloaded cd on drive %s\n", cdi->name);
			return unload_cd(vc);

		default:
			printk_debug("cdemu_ioctl: Unknown ioctl call 0x%x\n", cmd);
			return -EINVAL;
	}
}

static struct cdrom_device_ops cdemu_dops = {
	.open					= cdemu_open,
	.release				= cdemu_release,
	.drive_status			= cdemu_drive_status,
	.media_changed			= cdemu_media_changed,
	.tray_move				= cdemu_tray_move,
	.lock_door				= cdemu_lock_door,
	.select_speed			= cdemu_select_speed,
	.get_last_session		= cdemu_get_last_session,
	.get_mcn				= NULL,
	.reset					= cdemu_reset,
	.audio_ioctl			= cdemu_audio_ioctl,
	.dev_ioctl				= cdemu_ioctl,
	.capability				= CDC_OPEN_TRAY | CDC_LOCK |
							  CDC_MULTI_SESSION | CDC_MEDIA_CHANGED |
							  CDC_PLAY_AUDIO |
							  CDC_IOCTLS | CDC_DRIVE_STATUS,
	.n_minors				= 1,
};

#ifdef IS_LINUX_2_6
static int cdemu_block_open(struct inode *inode, struct file *file)
{
	printk_debug("cdemu_block_open: start\n");

	if (inode \
		&& inode->i_bdev \
		&& inode->i_bdev->bd_disk \
		&& inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
		printk_debug("cdemu_block_open: going into cdrom_open()\n");
		return cdrom_open(&vc->info, inode, file);
	}

	printk_debug("cdemu_block_open: -ENODEV\n");
	return -ENODEV;
}

static int cdemu_block_release(struct inode *inode, struct file *file)
{
	printk_debug("cdemu_block_release: start\n");

	if (inode \
		&& inode->i_bdev \
		&& inode->i_bdev->bd_disk \
		&& inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
		return cdrom_release(&vc->info, file);
	}

	printk_debug("cdemu_block_release: end\n");
	return -ENODEV;
}

static int cdemu_block_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
	printk_debug("cdemu_block_ioctl: start\n");

	if (inode \
		&& inode->i_bdev \
		&& inode->i_bdev->bd_disk \
		&& inode->i_bdev->bd_disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)inode->i_bdev->bd_disk->private_data;
		return cdrom_ioctl(&vc->info, inode, cmd, arg);
	}

	printk_debug("cdemu_block_ioctl: end\n");
	return -ENODEV;
}

static int cdemu_block_media_changed(struct gendisk *disk)
{
	printk_debug("cdemu_block_media_changed: start %s\n", disk->disk_name);

	if (disk && disk->private_data) {
		struct cdemu_device *vc = (struct cdemu_device *)disk->private_data;
		printk_debug("cdemu_block_media_changed: going into cdrom_media_changed()\n");
		return cdrom_media_changed(&vc->info);
	}

	printk_debug("cdemu_block_media_changed: -ENODEV\n");
	return -ENODEV;
}
#endif

static struct block_device_operations cdemu_bdops = {
#ifdef IS_LINUX_2_4
	.owner					= THIS_MODULE,
	.open					= cdrom_open,
	.release				= cdrom_release,
	.ioctl					= cdrom_ioctl,
	.check_media_change		= cdrom_media_changed,
#else
	.owner					= THIS_MODULE,
	.open					= cdemu_block_open,
	.release				= cdemu_block_release,
	.ioctl					= cdemu_block_ioctl,
	.media_changed			= cdemu_block_media_changed,
#endif
};

/* *********************************************************************** */
/*                        procfs code                                      */
/* *********************************************************************** */

#ifdef CONFIG_PROC_FS

#include <linux/proc_fs.h>
static struct proc_dir_entry* proc_entry;

static int proc_read(char* buf, char** start, off_t offset, int count, int* eof, void* data)
{
	int len = 0;
	int i;
	len += snprintf(len+buf, 1024, "Drive Loaded Comment\n");
	for (i=0; i < max_cdemus; i++) {
		if (cdemu_dev[i].loaded == 0)
			len += snprintf(len+buf, 1024, "%3i:     %i   NO_CD_LOADED\n", i,
				cdemu_dev[i].loaded);
		else
			len += snprintf(len+buf, 1024, "%3i:     %i   %s\n", i,
				cdemu_dev[i].loaded, cdemu_dev[i].comment);
	}
	*eof = 1;
	return len;
}

static int cdemu_proc_make(void)
{
	proc_entry = create_proc_entry(DEVICE_NAME, 0444, NULL);
	if (proc_entry)
		proc_entry->read_proc = proc_read;
	else
		return 0;
	return 1;
}

static void cdemu_proc_remove(void)
{
	if (proc_entry)
		remove_proc_entry("cdemu", NULL);
}

#else

static inline int cdemu_proc_make(void) {return 1;}
static inline void cdemu_proc_remove(void) {;}

#endif

/* *********************************************************************** */
/*                      load/unload code                                   */
/* *********************************************************************** */

int __init cdemu_init(void)
{
	int i;
	struct cdemu_device *vc;

	printk_debug("cdemu_init: start\n");

	/* check if the kernelparmeter is correct and register the block devices */
	#define	MIN_CDEMUS	1		/* should be self explanatory ... */
	#define	MAX_CDEMUS	255		/* 2.6 can have more than this but who would use it ? */
	if ((max_cdemus < MIN_CDEMUS) || (max_cdemus > MAX_CDEMUS)) {
		printk_cdemu(KERN_WARNING, "invalid max_cdemus (must be between %i and %i), using default (8)\n",
		             MIN_CDEMUS, MAX_CDEMUS);
		max_cdemus = 8;
	}

#ifdef IS_LINUX_2_6
	if (register_blkdev(MAJOR_NR, major_name)) {
#else
#ifdef CONFIG_DEVFS_FS
	if (devfs_register_blkdev(MAJOR_NR, "cdemu", &cdemu_bdops)) {
#else
	if (0) {
#endif
#endif
		printk_cdemu(KERN_ERR, "unable to register block device #%i for cdemu device\n", MAJOR_NR);
		return -EIO;
	}

	/* allocate max_cdemus entries in our struct */
	cdemu_dev = kmalloc(max_cdemus * sizeof(struct cdemu_device), GFP_KERNEL);
	if (!cdemu_dev) {
		printk_cdemu(KERN_ERR, "could not kmalloc enough memory for %i cdemu devices\n", max_cdemus);
		goto out_nomem;
	}

#ifdef IS_LINUX_2_4
	/* create the arrays for size and block size */
	cdemu_sizes = kmalloc(max_cdemus * sizeof(int), GFP_KERNEL);
	if (!cdemu_sizes)
		goto out_sizes;
	memset(cdemu_sizes, 0, max_cdemus * sizeof(int));

	cdemu_blksizes = kmalloc(max_cdemus * sizeof(int), GFP_KERNEL);
	if (!cdemu_blksizes)
		goto out_blksizes;
#endif

#ifdef IS_LINUX_2_6
#ifdef CONFIG_DEVFS_FS
	devfs_mk_dir("cdemu");
#endif
#endif

	for (i = 0; i < max_cdemus; ++i) {
		vc = &cdemu_dev[i];
		memset(vc, 0, sizeof(struct cdemu_device));
		vc->number = i;
		vc->loaded = 0;
		vc->media_changed = 0;
		vc->locked = 0;
		vc->pause = 0;
		vc->bdev = NULL;
		vc->info.ops = &cdemu_dops;
		vc->info.speed = 0; /* FIXME max_speed */
		vc->info.capacity = 1;
		vc->info.handle = vc;
		sprintf(vc->info.name, "%s%i", major_name, i);
		spin_lock_init(&vc->lock);
		printk_debug("cdemu_init: making cdrom #%i [%p]: %s\n", vc->number, vc, vc->info.name);

		/* register with the uniform cdrom driver */
#ifdef IS_LINUX_2_6
		vc->gd = alloc_disk(1);
		if (!vc->gd)
			goto out;
		vc->gd->major = MAJOR_NR;
		vc->gd->first_minor = i;
		sprintf(vc->gd->disk_name, "%s%i", major_name, i);
		vc->gd->fops = &cdemu_bdops;
		vc->gd->private_data = vc;
		set_capacity(vc->gd, 0);
		register_cdrom(&vc->info);
		vc->gd->queue = blk_init_queue(cdemu_request, &vc->lock);
		if (!vc->gd->queue)
			goto out;
		add_disk(vc->gd);
#ifdef CONFIG_DEVFS_FS
		devfs_mk_bdev(MKDEV(vc->gd->major, vc->gd->first_minor),
		              S_IFBLK|S_IRUSR|S_IWUSR,
		              "cdemu/%i", i);
#endif
#else
		vc->info.dev = MKDEV(MAJOR_NR, i);
		register_cdrom(&vc->info);
#ifdef CONFIG_DEVFS_FS
		devfs_plain_cdrom(&vc->info, &cdemu_bdops);
#endif
		cdemu_blksizes[i] = CD_FRAMESIZE;
#endif
	}

#ifdef IS_LINUX_2_4
#ifdef CONFIG_DEVFS_FS
	devfs_handle = devfs_mk_dir(NULL, "cdemu", NULL);
	devfs_register_series(devfs_handle, "%u", max_cdemus, DEVFS_FL_DEFAULT,
	                      MAJOR_NR, 0,
	                      S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP,
	                      &cdemu_bdops, NULL);
#endif

	/* tell the kernel our read/write function */
	blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), cdemu_request);

	blk_size[MAJOR_NR] = cdemu_sizes;
	blksize_size[MAJOR_NR] = cdemu_blksizes;
#endif

	if (!cdemu_proc_make()) {
		printk_debug("cdemu_init: Error registering /proc entry!\n");
		goto out_freeall;
	}

	/* all done report our success */
	printk_debug("cdemu_init: loaded (max %i devices)\n", max_cdemus);
	printk_debug("cdemu_init: end\n");
	printk_cdemu(KERN_INFO, "ver " CDEMU_MODULE_VERSION " loaded.  Registered %i cdemus.\n", max_cdemus); 
	return 0;

#ifdef IS_LINUX_2_6
out:
	for (i = 0; i < max_cdemus; ++i) {
		vc = &cdemu_dev[i];
		if (vc->gd->queue)
			blk_cleanup_queue(vc->gd->queue);
		if (vc->gd) {
			put_disk(vc->gd);
			del_gendisk(vc->gd);
		}
	}
#endif
out_freeall:
#ifdef IS_LINUX_2_4
	kfree(cdemu_blksizes);
out_blksizes:
	kfree(cdemu_sizes);
out_sizes:
#endif
	kfree(cdemu_dev);
out_nomem:
#ifdef IS_LINUX_2_4
#ifdef CONFIG_DEVFS_FS
	if (devfs_unregister_blkdev(MAJOR_NR, DEVICE_NAME))
#else
	if (0)
#endif
#else
	if (unregister_blkdev(MAJOR_NR, major_name))
#endif
		printk_cdemu(KERN_ERR, "unable to unregister block device\n");

	printk_cdemu(KERN_ERR, "cdemu_init: ran out of memory\n");
	return -ENOMEM;
}

void cdemu_exit(void)
{
	int i;
	struct cdemu_device *vc;

	printk_debug("cdemu_exit: start\n");

	/* get rid of the proc stuff */
	cdemu_proc_remove();

#ifdef IS_LINUX_2_4
	/* free all reference the kernel has in various caches */
	for (i = 0; i < max_cdemus; i++)
		fsync_dev(MKDEV(MAJOR_NR, i)); /* flush the devices */

	/* Fix up the request queue(s) */
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(MAJOR_NR));
#endif

	for (i = 0; i < max_cdemus; ++i) {
		vc = &cdemu_dev[i];
		printk_debug("cdemu_exit: unregistering cdrom #%i [%p]: %s\n", vc->number, vc, vc->info.name);
		/* this makes del_gendisk() error ?
		if (vc->gd->queue) {
			printk_debug("cdemu_exit: punting queue\n");
			blk_cleanup_queue(vc->gd->queue);
		}
		*/
#ifdef IS_LINUX_2_6
#ifdef CONFIG_DEVFS_FS
		devfs_remove("cdemu/%i", i);
#endif
		if (vc->bdev) {
			printk_debug("cdemu_exit: punting bdev\n");
			invalidate_bdev(vc->bdev, 1);
			bdput(vc->bdev);
		}
		printk_debug("cdemu_exit: punting gendisk\n");
		del_gendisk(vc->gd);
		put_disk(vc->gd);
#endif
		if (unregister_cdrom(&vc->info))
			printk(KERN_WARNING "cdemu: cannot unregister cdrom\n");
	}

#ifdef CONFIG_DEVFS_FS
#ifdef IS_LINUX_2_4
	devfs_unregister(devfs_handle);
#else
	devfs_remove("cdemu");
#endif
#endif

#ifdef IS_LINUX_2_4
#ifdef CONFIG_DEVFS_FS
	if (devfs_unregister_blkdev(MAJOR_NR, DEVICE_NAME))
#else
	if (0)
#endif
#else
	if (unregister_blkdev(MAJOR_NR, major_name))
#endif
		printk_cdemu(KERN_ERR, "unable to unregister block device\n");

	kfree(cdemu_dev);
#ifdef IS_LINUX_2_4
	kfree(cdemu_sizes);
	kfree(cdemu_blksizes);
#endif

	printk_debug("cdemu_exit: end\n");
	printk_cdemu(KERN_INFO, "ver " CDEMU_MODULE_VERSION " unloaded.  Unregistered %i cdemus.\n", max_cdemus); 
}

module_init(cdemu_init);
module_exit(cdemu_exit);

#ifndef MODULE
static int __init max_cdemus_setup(char *str)
{
	max_cdemus = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("max_cdemus=", max_cdemus_setup);
#endif

MODULE_AUTHOR("Robert Penz <robert.penz@outertech.com>, Mike Frysinger <vapier@netzero.com>");
MODULE_DESCRIPTION("CD Emulator for bin/cue images ver " CDEMU_MODULE_VERSION);
MODULE_LICENSE("GPL");
#ifdef IS_LINUX_2_6
module_param(max_cdemus, int, 0);
#else
MODULE_PARM(max_cdemus, "i");
#endif
MODULE_PARM_DESC(max_cdemus, "Maximum number of virtual cdroms (1-255)");
